##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""Customized HTML parser to find images in an imglist (similar to anchorlist)
"""

from htmllib import HTMLParser

class LCHTMLParser(HTMLParser):

    imglist = []

    def do_img(self, attrs):
        for attrname, value in attrs:
            if attrname == "src":
                if not self.imglist:
                    self.imglist = []
                self.imglist.append(value)



# vim:fileencoding=utf-8
# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""General utility functions."""

# Python imports
from urlparse import urlparse, urlunparse, urljoin
from sgmllib import SGMLParseError
import lxml.etree
import md5
import formatter
import StringIO

# CMF/Plone imports
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.public import RichWidget

# Zope imports
import gocept.linkchecker.log as log
from StructuredText import Basic
from StructuredText.DocumentClass import StructuredTextLink, DocumentClass

# Sibling imports
from gocept.linkchecker import lchtmllib


inet_services = {
    'ftp': 21,
    'http': 80,
    'https': 443,
    }


class urlrecord:
    """Representation of a URL with an attribute for each component.
    """

    components = ('scheme', 'netloc', 'path', 'parameters', 'query', 'fragment')

    def __init__(self, url):
        url_components = urlparse(url)

        for i, c in enumerate(self.components):
            setattr(self, c, url_components[i])

    def __str__(self):
        url_components = [getattr(self, c, '') for c in self.components]
        url = urlunparse(url_components)
        return url


def resolveRelativeLink(url, context):
    """Normalizes a URL according to RfC 2396 and throws away the fragment
    component, if any.
    """
    url_ = urlrecord(url)

    if url_.scheme:
        # Url has a protocol specified, e.g. HTTP, so we need not complete it.
        # We're only interested in a few protocols, though.
        if url_.scheme.lower() not in ['http', 'https', 'ftp']:
            return url
    elif url_.netloc:
        # A host is given, so we assume the HTTP protocol.
        url_.scheme = 'http'
    else:
        # We only have a path. We join it with the configured prefix (or use the
        # request if nothing was specified).
        lc = getToolByName(context, "portal_linkchecker")
        prefix = lc.database.defaultURLPrefix or context.REQUEST.BASE0
        if prefix.endswith('/'):
            prefix = prefix[:-1]
        portal_url = getToolByName(context, 'portal_url')
        prefix += '/%s' % portal_url.getRelativeContentURL(context)
        # Make up for a deviation of Python's urljoin from the test cases given
        # in RfC 2396, Appendix C.
        if url.startswith('?'):
            url = './' + url

        url = urljoin(prefix, url)
        url_ = urlrecord(url)

    url_.scheme = url_.scheme.lower()
    url_.netloc = url_.netloc.lower()

    if ':' in url_.netloc:
        hostname, port = url_.netloc.split(':')
        if port == str(inet_services.get(url_.scheme)):
            url_.netloc = hostname

    if not url_.path:
        url_.path = '/'

    url_.fragment = ''

    new_url = str(url_)
    return new_url


def retrieveSTX(text):
    """Retrieve links from STX data."""
    data = Basic(text)
    data = DocumentClass()(data)
    
    stack = data.getChildren()
    links = []
    while stack:
        element = stack.pop()
        if isinstance(element, StructuredTextLink):
            # Handle images seperately
            element = element.href
            if element.startswith('img:'):
                element = element[4:]
            links.append(element)
        try:
            stack.extend(element.getChildren())
        except AttributeError:
            pass

    return links


def retrieveHTML(text):
    parser = lchtmllib.LCHTMLParser(formatter.NullFormatter())
    try:
        parser.feed(text)
        parser.close()
    except SGMLParseError, e:
        # SGMLLib seems to die on bad HTML sometimes. (At least with python2.1)
        log.logger.error('retrieveHTML failed due to SGMLParseError: %s' % str(e))
    return tuple(parser.anchorlist) + tuple(parser.imglist)


def retrieveOneRichTextField(object, field):
    """
    Helper function for retrieveAllRichTextFields below and possibly
    useful for others.

    We would want to check something like
    (object.Schema()[fieldname].text_format
    but that attribute does not exist.
    Just try both retrieveSTX and retrieveHTML then.
    """
    text = field.getRaw(object)
    links = []

    for link in retrieveSTX(text):
        links.append(link)

    for link in retrieveHTML(text):
        links.append(link)
    return links


def retrieveAllRichTextFields(object):
    """Looks for all AT fields on an object that use the RichWidget
    and retrieves their links.
    """
    links = []
    for field in _getRichTextFields(object):
        for link in retrieveOneRichTextField(object, field):
            links.append(link)
    return links


## Update helpers

def updateAllRichTextFields(old_url, new_url, object):
    """Looks for all AT fields on an object that use the RichWidget
    and updates all occurances of the given url.
    """
    for field in _getRichTextFields(object):
        updateOneRichTextField(old_url, new_url, object, field)
    object.reindexObject()


def updateOneRichTextField(old_url, new_url, object, field):
    """Updates all occurances of old_url with new_url."""
    text = field.getRaw(object).decode('utf-8')
    # We can only handle HTML now.
    text = u"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
    <linkcheckerroot912ec803b2ce49e4a541068d495ab570>%s</linkcheckerroot912ec803b2ce49e4a541068d495ab570>""" % text
    parser = lxml.etree.HTMLParser()
    tree = lxml.etree.parse(StringIO.StringIO(text), parser)
    # links
    for link in tree.getiterator("a"):
        if link.get("href") == old_url:
            link.set("href", new_url)
    # images
    for image in tree.getiterator("img"):
        if image.get("src") == old_url:
            image.set("src", new_url)
    text = lxml.etree.tostring(tree, encoding=unicode)
    text = text.replace('\n<html xmlns="http://www.w3.org/1999/xhtml"><body><linkcheckerroot912ec803b2ce49e4a541068d495ab570>', '')
    text = text.replace('</linkcheckerroot912ec803b2ce49e4a541068d495ab570></body></html>', '')
    field.set(object, text)


def _getRichTextFields(object):
    return [f for f in object.Schema().fields()
              if isinstance(f.widget, RichWidget)]


def hash_link(link, object):
    """Provides a hash which can be used as an id for a Link object."""
    uid = object.UID()
    if link is None:
        link = ""
    if isinstance(link, unicode):
        link = link.encode('utf-8')
    return md5.md5("%s:%s" % (link, uid)).hexdigest()


def hash_url(url):
    return md5.md5(url).hexdigest()



# Copyright (c) 2005 gocept. All rights reserved.
# See also LICENSE.txt
# 
"""CMF link checker tool initalisation code"""

from gocept.linkchecker import permissions
from Products.CMFCore import utils as CMFutils
from Products.CMFCore import DirectoryView


def initialize_tool(context):
    from gocept.linkchecker import LinkCheckerTool
    tools = (LinkCheckerTool.LinkCheckerTool, )
    tool_init = CMFutils.ToolInit('Link checker tool',
                                  tools=tools,
                                  icon="tool.png")
    tool_init.initialize(context)
   
def initialize_content(context):
    from gocept.linkchecker import database, retrievemanager, reports
    context.registerClass(database.LinkDatabase,
                          constructors=(database.manage_addLinkDatabase,),
                          container_filter = filterLinkManagerAddable)
    context.registerClass(reports.BaseReports,
                          constructors=(reports.manage_addBaseReports,),
                          container_filter=filterLinkManagerAddable)

    context.registerClass(retrievemanager.RetrieveManager,
                          constructors=(retrievemanager.\
                                        manage_addRetrieveManager,),
                          container_filter=filterLinkManagerAddable
        )


def initialize(context):
    initialize_tool(context)
    initialize_content(context)
    DirectoryView.registerDirectory('skins', globals())
    

def filterLinkManagerAddable(objectmanager):
    """Tools for the Link Manager should not be addable everywhere"""
    if objectmanager.meta_type != "CMF Linkchecker Tool":
        return False
    return True



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
"""CMF link checker tool"""

# Python imports
import time
from types import StringTypes
from warnings import warn
import zope.interface

# Global Zope imports
from AccessControl import getSecurityManager, ClassSecurityInfo
from DateTime import DateTime
from OFS.Folder import Folder
from Globals import InitializeClass
from Acquisition import aq_base

# CMF/Plone imports
from Products.CMFCore.utils import UniqueObject, getToolByName
from Products.CMFCore.permissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression

from Products.Archetypes.interfaces.referenceable import IReferenceable

# Sibling imports
from gocept.linkchecker.interfaces import ILinkManager
from gocept.linkchecker.utils import resolveRelativeLink
from gocept.linkchecker import permissions, shorturls

LMS_REGISTRATION = "http://www.gocept.com/portal_lms_registration"


class LinkCheckerTool(UniqueObject, Folder):
    """A link checker tool. Encapsulating mechanisms for finding links in the
       site, testing their availability and notifying their owners."""

    zope.interface.implements(ILinkManager)

    id = 'portal_linkchecker'
    meta_type = 'CMF Linkchecker Tool'

    security = ClassSecurityInfo()

    plone_tool = 1

    manage_options = ( ({'label' : 'Overview', 'action' : 'manage_overview'},) +
                        Folder.manage_options )

    # Provide means to avoid retrieving/indexing/... events to trigger while
    # we're not set up correctly, yet.
    active = False

    #
    # ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = PageTemplateFile('www/explainLinkcheckerTool', globals(),
                                        __name__='manage_overview')

    def manage_afterAdd(self, item, container):
        LinkCheckerTool.inheritedAttribute('manage_afterAdd')(self, item,
                                                              container)
        object_ids = self.objectIds()

        if 'retrieving' not in object_ids:
            self.manage_addProduct["gocept.linkchecker"].\
                    manage_addRetrieveManager('retrieving')
        if 'database' not in object_ids:
            self.manage_addProduct["gocept.linkchecker"].\
                    manage_addLinkDatabase('database')
        if 'reports' not in object_ids:
            self.manage_addProduct["gocept.linkchecker"].\
                    manage_addBaseReports('reports')

        self.active = True

    ##############
    # ILinkManager

    security.declarePublic("shortURL")
    def shortURL(self, url, target=50):
        """Context aware wrapper for creating a shortened URL"""
        base = self.database.defaultURLPrefix
        return shorturls.shortURL(url, target, base)

    security.declarePublic("isManaged")
    def isManaged(self, object):
        """Tells if the object is managed by the link manager."""
        return self.retrieving.supportsRetrieving(object)

    security.declarePublic('resolveRelativeLink')
    def resolveRelativeLink(self, url, context):
        return resolveRelativeLink(url, context)

    def isUserAllowed(self):
        """returns whether the user is allowed to use Link Management services
        """
        security = getSecurityManager()
        portal = getToolByName(self, 'portal_url').getPortalObject()
        return security.checkPermission(permissions.USE_LINK_MANAGEMENT,
                                        portal)

    ######
    # misc

    security.declarePublic("registrationURL")
    def registrationURL(self):
        """Return the URL for registering with the gocept lms."""
        # XXX not sure which of REQUEST.URL or REQUEST.ACTUAL_URL is best here
        return_url = "%s#form_client_id" % self.REQUEST.URL 
        portal = getToolByName(self, 'portal_url').getPortalObject()
        callback_url = portal.portal_linkchecker.database.absolute_url()

        return "%s?callback_url=%s&return_url=%s" % (LMS_REGISTRATION,
                                                     callback_url, return_url)

    security.declareProtected(ManagePortal, 'getInfoFrameURL')
    def getInfoFrameURL(self):
        """Returns the URL received from the LMS or an empty string."""
        return self.database.getInfoFrameURL()


InitializeClass(LinkCheckerTool)



# Copyright (c) 2007 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: __init__.py 5173 2007-09-18 09:03:53Z sweh $

#namespace package boilerplate
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError, e:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)




# Copyright (c) 2004 gocept gmbh & co. kg
# See also LICENSE.txt
# 

import unittest

import time

from Products.Archetypes.tests.common import *
from Products.Archetypes.tests.utils import *

import Acquisition

from gocept.linkchecker.tests.base import LinkCheckerTestCase

from Products.CMFCore.utils import getToolByName

class Dummy(Acquisition.Implicit):
    
    started = False
    
    def start(self):
        """dummy method getting called by portal catalog"""
        self.started = True
        

class LinkCheckerReportTest(LinkCheckerTestCase):

    def test_ManagementOverview(self):
        lc = getToolByName(self.portal, 'portal_linkchecker')
        status = lc.reports.ManagementOverview()
    
        self.assertEquals(status.totalLinks, 0)
        self.assertEquals(status.links['red'], 0)
        self.assertEquals(status.links['grey'], 0)
        self.assertEquals(status.links['green'], 0)
        self.assertEquals(status.links['orange'], 0)
        self.assertEquals(status.linksPct['red'], 0)
        self.assertEquals(status.linksPct['grey'], 0)
        self.assertEquals(status.linksPct['green'], 0)
        self.assertEquals(status.linksPct['orange'], 0)
        

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LinkCheckerReportTest))
    return suite



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 

"""ShortURL tests



import unittest
from Testing import ZopeTestCase

from unittest import TestSuite, TestCase, makeSuite, main
from gocept.linkchecker.shorturls import shortURL


class ShortURLTest(TestCase):

    def testNormal(self):
        url = "http://www.gocept.com"
        self.assertEqual(shortURL(url), url)

        url = "http://www.gocept.com/test"
        self.assertEqual(shortURL(url, len(url)), url)

        url = "http://www.gocept.com/test/test1/test2;parameter?query=1#fragment"
        self.assertEqual(shortURL(url, len(url)), url)

    def testPathFolding(self):

        url = "http://www.gocept.com/angebot/opensource/gocept.linkchecker/revision1.3"
        self.assertEqual(shortURL(url), "http://www.gocept.com/angebot/.../revision1.3")
        self.assertEqual(shortURL(url, 40), "http://www.gocept.com/.../revision1.3")

    def testPriorityFolding(self):

        url = "http://www.gocept.com/angebot/opensource/gocept.linkchecker/revision1.3;asdf?query=1#fragment"
        
        self.assertEqual(shortURL(url, 89), 
              "http://www.gocept.com/angebot/opensource/gocept.linkchecker/revision1.3;asdf?query=1#...")

        self.assertEqual(shortURL(url, 84), 
              "http://www.gocept.com/angebot/opensource/gocept.linkchecker/revision1.3;asdf?...#...")
        
        self.assertEqual(shortURL(url, 60), 
              "http://www.gocept.com/angebot/.../revision1.3;...?...#...")


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ShortURLTest))
    return suite



from Products.Five import zcml
from Products.Five import fiveconfigure
from Testing import ZopeTestCase as ztc

from Products.Archetypes import public as atapi
from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import onsetup
from Products.PloneTestCase.setup import default_password, portal_owner


@onsetup
def setup_linkchecker_policy():
    ztc.installProduct('Five')

    fiveconfigure.debug_mode = True
    import gocept.linkchecker
    zcml.load_config('configure.zcml', gocept.linkchecker)
    fiveconfigure.debug_mode = False

    ztc.installPackage('gocept.linkchecker')
    ztc.installProduct('ZCatalog')


setup_linkchecker_policy()
ptc.setupPloneSite(products=['gocept.linkchecker'])


class LinkCheckerTestCase(ptc.FunctionalTestCase):

    def __init__(self, *args, **kw):
        super(LinkCheckerTestCase, self).__init__(*args, **kw)
        self.portal_owner = portal_owner
        self.default_password = default_password



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""CMF link checker tool - url object
"""

# Zope imports
from DateTime import DateTime
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

# CMF/Plone imports
from Products.CMFCore.utils import getToolByName

# Sibling imports
import gocept.linkchecker.interfaces
import gocept.linkchecker.utils


class URL(SimpleItem):
    """URL status information"""

    meta_type = "URL"

    __implements__ = (gocept.linkchecker.interfaces.IURL,)
    # XXX make accessor functions
    __allow_access_to_unprotected_subobjects__ = 1

    url = None
    lastcheck = None
    lastupdate = None
    laststate = "grey"
    state = "grey"
    reason = ""
    registered = False

    manage_options = ({'label': 'Info', 'action':'manage_info'},) + \
                     SimpleItem.manage_options

    manage_info = PageTemplateFile('www/url', globals(), __name__='manage_info')

    def __init__(self, url):
        self.id = gocept.linkchecker.utils.hash_url(url)
        self.url = url

    def getLinks(self):
        lc = getToolByName(self, "portal_linkchecker")
        return lc.database.getLinksForURL(self.url)

    def updateStatus(self, state, reason):
        assert state in ['red', 'green', 'orange', 'grey', 'blue'], \
            "Invalid state %s" % state
        self.reason = reason
        now = DateTime()
        self.lastcheck = now
        if state != self.state:
            self.laststate = self.state
            self.state = state 
            self.lastupdate = now
        self.index()
        # Reindex link objects to update their status caches
        for link in self.getLinks():
            link.index()

    # Python/Zope helper functions
    def __repr__(self):
        return "<URL %s (%s)>" % (self.url, self.state)

    def _p_resolveConflict(self, oldState, savedState, newState):
        # This happens if the LMS starts talking to us and we do crawling
        # and the site is busy in general.
        # However, we are only interested in the news information we get,
        # so we can happily do something simple here.
        return newState

    # Catalog support
    def manage_afterAdd(self, container, object):
        self.index()

    def manage_beforeDelete(self, item, container):
        return
        self.unindex()

    def index(self):
        path = '/'.join(self.getPhysicalPath())
        self.url_catalog.catalog_object(self, path)

    def unindex(self):
        path = '/'.join(self.getPhysicalPath())
        self.url_catalog.uncatalog_object(path)



# Copyright (c) 2003-2009 gocept gmbh & co. kg
# See also LICENSE.txt
"""gocept.linkchecker - link information object"""

from DateTime import DateTime
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.utils import getToolByName
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import gocept.linkchecker.interfaces
import gocept.linkchecker.utils
import zope.interface 


class Link(SimpleItem):
    """A link as used from an object."""

    meta_type = "Link"

    zope.interface.implements(gocept.linkchecker.interfaces.ILink)

    __allow_access_to_unprotected_subobjects__ = 1

    manage_options = ({'label': 'Info', 'action':'manage_info'},) + \
                     SimpleItem.manage_options
    manage_info = PageTemplateFile('www/link', globals(), __name__='manage_info')

    def __init__(self, link, id, object):
        self.object = object
        self.link = link
        self.id = id

    def __setstate__(self, state):
        for key in ['url', 'state', 'lastcheck', 'reason']:
            if 'url' in state:
                del state['url']
        SimpleItem.__setstate__(self, state)

    # gocept.linkchecker.interfaces.ILink
    def getURL(self):
        """Return the URL object this link refers to."""
        lc = getToolByName(self, "portal_linkchecker")
        url = self.url()
        res = lc.database.queryURLs(url=url)
        if not res:
            # Ensure the URL exists
            raise ValueError("Could not find URL %s." % url)
        return res[0].getObject()

    def getObject(self):
        """Return a reference to the object."""
        references = getToolByName(self, "reference_catalog")
        return references.lookupObject(self.object)

    # Python/Zope helpers 
    def __repr__(self):
        return "<Link %s at %s>" % (self.link, self.object)

    # Catalog support
    def manage_afterAdd(self, container, object):
        self.index()

    def manage_beforeDelete(self, item, container):
        self.unindex()

    def index(self):
        if self.getObject() is None:
            self.getParentNode().manage_delObjects([self.id])
            return 
        try:
            url = self.getURL()
        except ValueError:
            # Make sure the URL object exists when we index ourselves.
            url = gocept.linkchecker.url.URL(self.url())
            db = self.getLinkCheckerDatabase()
            if url.id not in db.objectIds():
                db._setObject(url.id, url)
            url = db[url.id]
        path = '/'.join(self.getPhysicalPath())
        self.link_catalog.catalog_object(self, path)

    def unindex(self):
        path = '/'.join(self.getPhysicalPath())
        try:
            catalog = self.link_catalog
        except AttributeError:
            # This is a special case to handle very large databases: empty or
            # delete the catalog then delete stuff
            return
        catalog.uncatalog_object(path)

    def url(self):
        return gocept.linkchecker.utils.resolveRelativeLink(
            self.link, self.getObject())

    def state(self):
        return self.getURL().state

    def reason(self):
        return self.getURL().reason

    def lastcheck(self):
        return self.getURL().lastcheck



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""gocept.linkchecker interface definitions"""


from zope.interface import Interface, Attribute

from Products.Archetypes import public as atapi


class ILinkManager(Interface):

    database = Attribute("A link database implementing the "
                         "ILinkDatabase interface.")
    reports = Attribute("A report utility implementing the "
                        "IBaseReports interface.")
    retrieving = Attribute("A retrieve manager implementing the "
                           "IRetrieveManager interface.")

    def shortURL(url, target=50):
        """Context aware wrapper for creating a shortened URL"""

    def isManaged(object):
        """Tells if the object is managed by the link manager."""

    def resolveRelativeLink(url, context):
        """Resolve relative URL from context to an absolute URL.

        This is similar to the absolute_url(), but does not require an object
        and honours the prefix settings from the link management settings.
        """

    def isUserAllowed():
        """Returns whether the user is allowed to use link management
        services.

        Returns True, iff the currently logged in user is allowed to use link
            management services at all.

        This function checks whether the user has the 'Use link management functions'
        permission in the Plone root.
        """

    def getLinkManager():
        """return self"""


class IBaseReports(Interface):
    """Supplement functions for various basic reports."""

    def GroupedLinksForAuthenticatedMember():
        """Returns a list of links aggregated by state for the current user.

           The list only features catalog brains from the database.
        """

    def DocumentsInState(state):
        """Returns a list of all objects that contain links in the given state.
        """

    def LinksInState(state):
        """Returns a list of links in the given state."""

    def ManagementOverview():
        """Returns a comprehensive management overview over the complete site.
        """


class ILinkDatabase(Interface):
    """A database of links. 

       Manages the physical storage of link information.
    """

    defaultURLPrefix = Attribute("A valid URL that is used (component wise) to "
                                 "prepend incomplete links")

    def configure(defaultURLPrefix=None):
        """Set configuration values."""

    def unregisterLink(link, object):
        """Unregisters a given link/object pair at the database.

           If the given link isn't referenced by any object anymore, it will 
           be removed from the database completely.

           Raises KeyError if the given pair isn't registered.

           Returns None.
        """

    def unregisterObject(object):
        """Unregisters all links for this object.

           Returns None.
        """

    def getLinksForObject(object):
        """Retrieve information about an object.

           Returns a list of ILink objects. 
           Returns an empty list if the object hasn't been registered yet.
        """

    def getLinksForURL(url):
        """Retrieve links for an url.

        Returns all links that point to the given url.
        """

    def getAllLinks():
        """Returns a list of all ILink objects.

           Warning: This is likely to be a slow method in large data sets.
        """

    def getAllLinkIds():
        """Returns a list of the ids of all ILink objects.

           Warning: This is likely to be a slow method in large data sets.
        """

    def getLinkCount():
        """Returns the amount of links currently in the database."""

    def queryLinks(self, **args):
        """Returns the result of querying the ZCatalog that indexes links."""

    def queryURLs(self, **args):
        """Returns the result of querying the ZCatalog that indexes urls."""


class IRetrieveManager(Interface):
    """Utility to perform link retrieval."""

    def retrieveObject(object):
        """Retrieves the links from the given object."""

    def retrieveSite():
        """Retrieves the links from all objects in the site.

           Raises RuntimeError if no connection to the lms can be established.
        """

    def supportsRetrieving(object):
        """Tells if the object is supported for retrieving links."""

    def getAllRetrieverNames():
        """Returns the name of all available retrievers."""


class IGlobalRetrieverRegistry(Interface):
    """Process global registry that holds information
       about all system wide available retrievers.
    """

    def register(method, name, defaults=[]):
        """Register a retriever method globally under the given title and as a
           default for the given portal_types.

           Defaults are overwritten subsequently by later calls.

           Raises ValueError if the method has been registered already.

           Returns None
        """

    def getByName(name):
        """Returns a retriever by it's name.

           Raises KeyError if a retriever with this name doesn't exists.
        """

    def getDefault(type):
        """Returns the default retriever for a type.

           Raises KeyError if there is no retriever for the given type.
        """

    def listNames():
        """Returns a list of all retriever names."""


class IURL(Interface):
    """URL status information

    The URL as transmitted to the LMS. It must be absolute and include protocol
    and netlocation.

    A URL can be referred to by many links as, depending on the context, a URL
    can be expressed by several relative spellings.

    """

    url = Attribute("Contains the URL the link refers to.")
    registered = Attribute("Has this link been registered with the web service?")
    lastcheck = Attribute("DateTime when the last check was performed.") 
    lastupdate = Attribute("DateTime when the last status change ocurred.")
    laststate = Attribute("The state before the current state was assumed.")
    state = Attribute("The current state of the link. Can be one of ['red', " \
                      "'green', 'orange', 'blue', 'grey'].")
    reason = Attribute("A verbose reason describing the error")
    links = Attribute(
        "A property that returns all currently associated links"
        "(as catalog brains).")

    def updateStatus(state, reason):
        """Updates the status information for this URL."""


class ILink(Interface):
    """A link as used from an object."""

    link = Attribute("Link as stored on the object")
    object = Attribute("The UID of the object containing the link.")
    url = Attribute(
        "URL as computed from the link and the object as context.")

    # Fields to support caching of data
    state = Attribute("The current state of the URL.")
    reason = Attribute("The current state of the URL.")

    def getObject():
        """Returns a list of objects referencing this link."""

    def getURL():
        """Return the URL object this link refers to."""


class IRetriever(Interface):
    """A class that implements a routine to retrieve and update URLs stored on
    an object.
    """

    def retrieveLinks():
        """Finds all links from the object and return them.

           Returns a list of URLs.

        """

    def updateLink(oldurl, newurl):
        """Replace all occurances of <oldurl> on object with <newurl>."""



##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool interface definitions"""



from Products.CMFCore.permissions import *

RUN_MAIL_NOTIFICATIONS = "Linkchecker: Run mail notifications"
USE_LINK_MANAGEMENT = "Linkchecker: Use link management functions"

# XXX initialize the permissions with default roles
#    MAIL -> manager
# USE_LINK_MANAGEMENT -> member



# Copyright (c) 2003-2006 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""CMF link checker tool - link database"""

from AccessControl import getSecurityManager, ClassSecurityInfo
from Globals import InitializeClass
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
from Products.CMFCore import permissions
from gocept.linkchecker import permissions
from gocept.linkchecker.interfaces import ILinkDatabase
from xmlrpclib import ServerProxy, Fault
from zExceptions import Unauthorized
import Products.Archetypes.interfaces
import Products.CMFCore.utils
import gocept.linkchecker.link
import gocept.linkchecker.log as log
import gocept.linkchecker.url
import gocept.linkchecker.utils
import socket
import zope.interface


PROTOCOL_VERSION = 2
WEBSERVICE = "http://lms.gocept.com/v2"

WEBSERVICE_STATEMAP = {
    "unknown" : "grey",
    "unsupported protocol": "blue",
    "ok" : "green",
    "temporary unavailable": "orange",
    "unavailable": "red"}


class OfflineWebserviceDummy(object):
    """An offline dummy for the web service."""

    def __init__(self):
        self.set_notifications = []
        self.register_many = []
        self.unregister_many = []

    def setClientNotifications(self, option):
        self.set_notifications.append(option)

    def registerManyLinks(self, links):
        self.register_many.append(links)
        return []

    def unregisterManyLinks(self, links):
        self.unregister_many.append(links)

offline_webservice = OfflineWebserviceDummy()


def manage_addLinkDatabase(container, id):
    """Add a new link database to a link manager."""
    container._setObject(id, LinkDatabase(id))


class LinkDatabase(BTreeFolder2):
    """A database of links. 

    Manages the physical storage of link and url information.
    """

    zope.interface.implements(ILinkDatabase)

    defaultURLPrefix = ""

    webservice = WEBSERVICE
    clientid = "change_me"
    password = ""

    offline = False

    security = ClassSecurityInfo()

    def getLinkCheckerDatabase(self):
        """Retrieve this object with a meaningful name via acquisition."""
        return self.aq_inner

    def manage_afterAdd(self, item, container):
        LinkDatabase.inheritedAttribute('manage_afterAdd')(self, item,
                                                           container)

        def insertIndexes(catalog, indexes):
            existing = catalog.Indexes.objectIds()
            for index_name, index_type in indexes:
                if index_name not in existing:
                    catalog.addIndex(index_name, index_type)

        def insertCacheColumns(catalog, columns):
            existing_columns = catalog.schema()
            for c, value in columns:
                if c not in existing_columns:
                    catalog.addColumn(c, value)

        # Setup the link catalog
        if 'link_catalog' not in self.objectIds():
            self.manage_addProduct['ZCatalog'].\
                    manage_addZCatalog('link_catalog', 'Link catalog')
        indexes = [('state', 'KeywordIndex'),
                   ('url', 'FieldIndex'),
                   ('object', 'KeywordIndex')]
        insertIndexes(self.link_catalog, indexes)

        columns = [("object", None),
                   ("url", ""),
                   ("reason", ""),
                   ("lastcheck", ""),
                   ("state", ""),
                   ("link", ""),
                   ("getId", "")]
        insertCacheColumns(self.link_catalog, columns)

        # Setup the url catalog
        if 'url_catalog' not in self.objectIds():
            self.manage_addProduct['ZCatalog'].\
                    manage_addZCatalog('url_catalog', 'URL catalog')
        indexes = [('url', 'FieldIndex'),
                   ('registered', 'FieldIndex')]
        insertIndexes(self.url_catalog, indexes)

        columns = [("url", ""),
                   ("id", ""),
                   ("reason", ""),
                   ("lastcheck", ""),
                   ("state", "grey")]
        insertCacheColumns(self.url_catalog, columns)

    ###############
    # ILinkDatabase

    security.declareProtected(permissions.ManagePortal, 'configure')
    def configure(self, defaultURLPrefix=None, clientid=None, password=None,
                  webservice=None, client_notifications=None):
        """Set configuration values."""
        if clientid is not None:
            self.clientid = clientid
        if password:
            self.password = password
        if defaultURLPrefix is not None:
            self.defaultURLPrefix = defaultURLPrefix
        if webservice is not None:
            self.webservice = webservice

        # need to reconfigure connection
        try:
            delattr(self, '_v_lms_client')
        except AttributeError:
            pass

        if client_notifications is not None:
            server = self._getWebServiceConnection()
            if server is not None:
                server.setClientNotifications(client_notifications)

    security.declareProtected(
        permissions.USE_LINK_MANAGEMENT, 'registerLinks')
    def registerLinks(self, links, object, online=True):
        """Registers links for an object at the database."""
        sm = getSecurityManager()
        if not sm.checkPermission(
            permissions.ModifyPortalContent, object):
            raise Unauthorized, \
                "Can't modify link registrations for this object."
        link_objects = [self._register_link(link, object) for link in links]
        if online:
            urls = [link.getURL()
                    for link in link_objects
                    if link is not None]
            self._register_urls_at_lms(urls)

    def _register_link(self, link, object):
        link_id = gocept.linkchecker.utils.hash_link(link, object)
        if link_id in self.objectIds():
            return
        link = gocept.linkchecker.link.Link(link, link_id, object.UID())
        # Now we can add the link to the database 
        self._setObject(link_id, link)
        return self[link_id]

    security.declarePrivate("_updateWSRegistrations")
    def _updateWSRegistrations(self):
        """Registers a link with the webservice. 

        May also register yet non-registered other links.
        """
        unregistered = self.queryURLs(registered=False)
        unregistered = [x.getObject() for x in unregistered]
        # Nothing to do?
        if not unregistered:
            return
        self._register_urls_at_lms(unregistered)

    def _register_urls_at_lms(self, url_objects):
        """Register the given URL objects at the LMS web service.
        """
        # Do we have an LMS connection?
        lms = self._getWebServiceConnection()
        if lms is None:
            return
        urls = [url.url for url in url_objects]
        states = lms.registerManyLinks(urls)
        # The server *may* report the status of URls it already knows.
        for url, state, reason in states:
            state = WEBSERVICE_STATEMAP[state]
            url = self[gocept.linkchecker.utils.hash_url(url)]
            url.updateStatus(state, reason)
        # Mark all of the URLs as registered
        for url in url_objects:
            url.registered = True

    security.declarePrivate('unregisterLink')
    def unregisterLink(self, link, object):
        """Unregisters a given link/object pair at the database.

           If the given link isn't referenced by any object anymore, it will 
           be removed from the database completely.

           Returns None.
        """
        id = gocept.linkchecker.utils.hash_link(link)
        self.manage_delObject(id)

    security.declarePrivate('unregisterObject')
    def unregisterObject(self, object):
        """Unregisters all links for this object.

           Returns None.
        """
        if isinstance(object, (gocept.linkchecker.url.URL,
                               gocept.linkchecker.link.Link)):
            return
        links = self.getLinksForObject(object)
        link_ids = [x.getId() for x in links]
        self.manage_delObjects(link_ids)

    def _getWebServiceConnection(self):
        """return connection to lms

        returns LMSClient instance if connection is ok
        returns None if connection is not Ok
        """
        conn = self.checkConnection()
        if conn.ok:
            client = self._get_connection_object()
        else:
            log.logger.error('Connection to LMS not possible: %s' % (conn.error, ))
            if self.offline:
                client = offline_webservice
            else:
                client = None
        return client

    def _get_connection_object(self):
        client = getattr(self, '_v_lms_client', None)
        if client is None:
            client = LMSClient(self.webservice, self.clientid, self.password)
            self._v_lms_client = client
        return client

    def _deleteURLs(self, urls):
        # Unregister with LMS
        lms = self._getWebServiceConnection()
        if lms is None:
            return
        self.manage_delObjects([url.id for url in urls])
        lms.unregisterManyLinks([url.url for url in urls])

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getLinksForURL')
    def getLinksForURL(self, url):
        """Retrieve links for an url."""
        links = [x.getObject() for x in self.queryLinks(url=url)]
        links = filter(None, links)
        return links

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getLinksForObject')
    def getLinksForObject(self, object, state=None):
        """Retrieve information about an object.

           Returns a list of ILink objects.
           Returns an empty list if the object hasn't been registered yet.
        """
        sm = getSecurityManager()
        if not sm.checkPermission(
            permissions.AccessContentsInformation, object):
            raise Unauthorized, "Can't view links on this object."
        if not Products.Archetypes.interfaces.IReferenceable.providedBy(object):
            return []

        uid = object.UID()
        if uid is None:
            links = []
        else:
            query_args = {}
            if state is not None:
                query_args['state' ] = state
            links = self.queryLinks(object=[uid], **query_args)
            links = [x.getObject() for x in links]
            links = filter(None, links)
        return links

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getAllLinks')
    def getAllLinks(self):
        """Returns a list of all ILink objects.

           Warning: This is likely to be a slow method in large data sets.
        """
        return self.objectValues(['Link'])

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getAllLinkIds')
    def getAllLinkIds(self):
        """Returns a list of the ids of all ILink objects.

           Warning: This is likely to be a slow method in large data sets.
        """
        return self.objectIds(['Link'])

    security.declareProtected(
        permissions.USE_LINK_MANAGEMENT, 'getLinkIterator')
    def getLinkIterator(self):
        """returns generator for all link objects"""
        for id in self.getAllLinkIds():
            yield self.get(id)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getLinkCount')
    def getLinkCount(self):
        """Returns the amount of links currently in the database."""
        return len(self.queryLinks())

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'queryLinks')
    def queryLinks(self, **args):
        """Returns the result of querying the ZCatalog that indexes links."""
        return self.link_catalog(**args)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'queryURLs')
    def queryURLs(self, **args):
        """Returns the result of querying the ZCatalog that indexes urls."""
        return self.url_catalog(**args)

    security.declareProtected(
        permissions.ManagePortal, 'sync')
    def sync(self):
        lms = self._getWebServiceConnection()
        lms.forceSynchronization()

    security.declareProtected(
        permissions.ManagePortal, 'updateLinkStatus')
    def updateLinkStatus(self, url, state, reason):
        """XML-RPC connector for LMS"""
        state = WEBSERVICE_STATEMAP[state]
        for url in self.queryURLs(url=url):
            url = url.getObject()
            url.updateStatus(state, reason)

    security.declarePublic('updateManyStates')
    def updateManyStates(self, client_id, password, update_list):
        """XML-RPC connector for LMS"""
        self.authenticateXMLRPC(client_id, password)
        for url, state, reason in update_list:
            self.updateLinkStatus(url, state, reason)

    security.declarePublic('xmlrpc_getAllLinks')
    def xmlrpc_getAllLinks(self, client_id, password):
        """XML-RPC connector for LMS

           returns all links for LMS
        """
        self.authenticateXMLRPC(client_id, password)
        # XXX optimiziation: make .url unique
        return [x.url for x in self.queryURLs()]

    security.declarePrivate('authenticateXMLRPC')
    def authenticateXMLRPC(self, client_id, password):
        """authenticates XML remote call"""
        if not (client_id == self.clientid and\
                password == self.password):
            raise Unauthorized,\
                "Client-Id or password do not match."

    security.declareProtected(
        permissions.ManagePortal, 'updateAllStatus')
    def updateAllStatus(self):
        """Manually update all link status'
        """
        server = self._getWebServiceConnection()
        if server is None:
            return
        for url in self.queryURLs():
            try:
                status, reason = server.getStatus(url.url)
            except KeyError:
                # not registered with server
                server.registerLink(url.url)
                status = 'temporary unavailable'
                reason = 'No status response from LMS yet.'
            status = WEBSERVICE_STATEMAP[status]
            if url.state != status:
                url = url.getObject()
                url.updateStatus(status, reason)

    security.declareProtected(
        permissions.USE_LINK_MANAGEMENT, 'checkConnection')
    def checkConnection(self):
        """return if connection is up and running

            returns r.ok == True if connection is ok
            returns r.ok == False if connection is *not* ok
                sets also r.error to some error information
        """
        class R:
            __allow_access_to_unprotected_subobjects__ = 1
            def __init__(self, **kwargs):
                self.__dict__.update(kwargs)

        try:
            server = self._get_connection_object()
        except Exception, e:
            ok = False
            error = '%s: %s' % (str(e.__class__), str(e))
            protocol_version='unknown',
        else:
            ok, protocol_version, error = server.checkConnection()
        r = R(ok=ok, error=error,
              client_protocol=PROTOCOL_VERSION,
              server_protocol=protocol_version)
        return r

    security.declareProtected(
        permissions.ManagePortal, 'getClientNotifications')
    def getClientNotifications(self):
        server = self._getWebServiceConnection()
        if server is None:
            return False
        stat = server.getClientNotifications()
        return stat

    security.declareProtected(permissions.ManagePortal, 
                              'getInfoFrameURL')
    def getInfoFrameURL(self):
        """Returns the URL received from the LMS or an empty string."""
        server = self._getWebServiceConnection()
        if server is None:
            return ""
        return server.getInfoFrameURL()


class LMSClient:
    """client to lms"""

    _exception_mapping = {
        'KeyError': KeyError,
    }

    def __init__(self, url, client_id, password):
        self.url = url
        self.client_id = client_id
        self.password = password
        self._server = ServerProxy(url)

    def forceSynchronization(self):
        self._server.forceSynchronization(self.client_id, self.password)

    def registerManyLinks(self, urls):
        result = []
        if urls:
            try:
                result = self._server.registerURLs(self.client_id, self.password, urls)
            except Fault, f:
                self._translate_fault(f)
        return result

    def unregisterManyLinks(self, urls):
        if not urls:
            return []
        try:
            result = self._server.unregisterURLs(self.client_id, self.password, urls)
        except Fault, f:
            self._translate_fault(f)
        return result

    def getStatus(self, url):
        try:
            status = self._server.getStatus(url)
        except Fault, f:
            self._translate_fault(f)
        return status

    def getClientNotifications(self):
        try:
            return self._server.getClientNotifications(self.client_id, self.password)
        except Fault, f:
            self._translate_fault(f)

    def getInfoFrameURL(self):
        try:
            return self._server.getInfoFrameURL(self.client_id, self.password)
        except Fault, f:
            self._translate_fault(f)

    def setClientNotifications(self, status):
        if isinstance(status, (str, unicode)):
            status = int(status)
        if isinstance(status, int):
            status = bool(status)
        self._server.setClientNotifications(self.client_id, self.password, status)

    def checkConnection(self):
        is_ok = False
        error = 'unknown'

        protocol_version = None
        try:
            protocol_version = self._server.checkConnection(self.client_id, self.password)
        except Fault, f:
            error = '%s (%s)' % (f.faultString, f.faultCode)
        except socket.error, e:
            e_code, e_string = e.args
            error = 'socket.error: %s (%s)' % (e_string, e_code)
        except Exception, e:
            error = str(e)
        else:
            if protocol_version == PROTOCOL_VERSION:
               is_ok = True 
            else:
               error = ("Incompatible protocol version. " 
                        "Got: %s, expected: %s" % (protocol_version,
                                                   PROTOCOL_VERSION))
        return is_ok, protocol_version, error

    #########
    # private

    def _translate_fault(self, f):
        name, value = f.faultCode, f.faultString

        exception_class = self._exception_mapping.get(name)
        if exception_class is None:
            sm = zope.component.getSiteManager()
            Products.CMFCore.utils.getToolByName(
                sm, "plone_utils").addPortalMessage(
                    "LMS Error: %s/%s" % (name, value), type='error')
        else:
            e = exception_class(value)
            raise e


InitializeClass(LinkDatabase)



# Copyright (c) 2004 gocept gmbh & co. kg
# See also LICENSE.txt
# 

import Acquisition
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.tests.test_sitepolicy import makeContent
from gocept.linkchecker.tests.base import LinkCheckerTestCase
from gocept.linkchecker.interfaces import ILink

import time
import unittest
import zope.lifecycleevent


class LinkCheckerTest(LinkCheckerTestCase):

    def test_link_registration(self):
        self.loginAsPortalOwner()
        doc = makeContent(self.portal, portal_type="Document", id="asdf")

        lc = getToolByName(self.portal, 'portal_linkchecker')
        links = ["http://www.gocept.com", 
                 "http://www.asfdhasdkjfhasdkjhfkjsda.org", 
                 "http://www.zope.org", "http://www.microsoft.com"]
        info = lc.database.registerLinks(links, doc)

        # Check for the links the object was registered for
        doc_links = lc.database.getLinksForObject(doc)
        self.assertEqual(len(links), len(doc_links))

        for l in doc_links:
            self.failUnless(l.link in links)

    def test_crawling(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')
        lc['database'].offline = True
        
        links = ["http://www.google.de/", "http://www.gocept.com/",
            "http://www.arcor.de/", "http://www.gocept.com/"]
        docs = []
        
        for x in links:
            id = self.portal.generateUniqueId()
            doc = makeContent(self.portal, portal_type="Document", id=id)
            doc.edit('html', '<a href="%s">asdf</a>' % x)
            docs.append(doc)

        lc.retrieving.retrieveSite()

        # XXX 25 is an arbitrary amount as the stock Plone site already
        # contains an unknown number.
        self.assertEqual(lc.database.getLinkCount(), 25)

        # However, the three URLs provided above are now known and have
        # associated links:
        google = lc.database.getLinksForURL('http://www.google.de/')
        self.assertEquals(len(google), 1)
        gocept = lc.database.getLinksForURL('http://www.gocept.com/')
        self.assertEquals(len(gocept), 2)
        arcor = lc.database.getLinksForURL('http://www.arcor.de/')
        self.assertEquals(len(arcor), 1)

        # Each of the documents that we created has a link associated now
        # which is in the list of the URLs above.
        for doc in docs:
            l = lc.database.getLinksForObject(doc)
            self.assertEqual(len(l), 1)
            self.failUnless(ILink.providedBy(l[0]))
            self.failUnless(l[0].url in links)

    def test_deletion(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')
        lc['database'].offline = True
        lc.retrieving.retrieveSite()
        # XXX 21 is an arbitrary amount as the stock Plone site already
        # contains an unknown number.
        self.assertEqual(21, lc.database.getLinkCount())

        # Now we delete an object and re-crawl the site. This should cause the
        # number of links to drop.
        self.portal.manage_delObjects(['front-page'])
        self.assertEqual(0, lc.database.getLinkCount())

    def test_crawl_offline(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')
        lc.database.configure(clientid='does-not-exist')

    def test_sync_failing(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')
        self.assertRaises(AttributeError, lc.database.sync)

    def test_byobject(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')
        db = lc.database
        doc = makeContent(self.portal, portal_type="Document", id="asdf")
        doc.setText('''"<a href='bla/fasel'>foo</a>
                    <a href='http://www.gocept.com'>gocept</a>''')
        
        lc.retrieving.retrieveObject(doc)

        links = db.getLinksForObject(doc)
        self.assertEquals(2, len(links))
     
    def test_atsupport(self):
        self.loginAsPortalOwner()
        portal = self.portal
        lc = getToolByName(portal, 'portal_linkchecker')
        db = lc.database
        portal.invokeFactory('Document', 'd1')
        d1 = portal.d1
        
        d1.setText('<a href="http://www.gocept.com">bla</a>')
        zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(d1))
        links = db.getLinksForObject(d1)
        self.assertEquals(1, len(links))

        d1.setText('foo')
        zope.event.notify(zope.lifecycleevent.ObjectModifiedEvent(d1))
        links = db.getLinksForObject(d1)
        self.assertEquals(0, len(links))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LinkCheckerTest))
    return suite



# vim:fileencoding=utf-8
# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id$

import unittest
from gocept.linkchecker.tests.base import LinkCheckerTestCase
from Products.Five.testbrowser import Browser

class TestSetup(LinkCheckerTestCase):

    def test_portal_tool(self):
        self.assert_('portal_linkchecker' in self.portal.objectIds())

    def test_actions(self):
        # Ensure that all actions were registered.
        b = Browser()
        b.open(self.portal.absolute_url()+'/login')
        b.getControl(name='__ac_name').value = self.portal_owner
        b.getControl(name='__ac_password').value = self.default_password
        b.getControl(name='submit').click()
        b.open(self.portal.absolute_url())
        self.assert_(b.getLink('Link management'))
        self.assert_(b.getLink('My links'))
        self.assert_(b.getLink('Links'))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSetup))
    return suite



# Copyright (c) 2004-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 

import unittest

from Products.Archetypes.tests.common import *
from Products.Archetypes.tests.utils import *

from gocept.linkchecker.tests.base import LinkCheckerTestCase

from Products.CMFCore.utils import getToolByName


class LinkCheckerTest(LinkCheckerTestCase):

    def test_normalization(self):
        # Test if relative link will get absolute in Database
        self.loginAsPortalOwner()
        portal = self.portal
        pl = getToolByName(portal, 'portal_linkchecker')
	pl.database.defaultURLPrefix = portal.absolute_url()

        portal.invokeFactory('Document', 'doc1')
        
        doc1 = portal.doc1
        doc1.setText('bla<a href="index_html">index.html</a>\nbumm')
        pl.retrieving.retrieveObject(doc1)
        db_cont = pl.database.queryLinks()
        self.assertEqual(1, len(db_cont))
        self.assertEqual(portal.absolute_url() + "/index_html", db_cont[0].url)

        
        ## assert '..' are stripped
    
        portal.invokeFactory('Folder', 'f1')
        portal.invokeFactory('Folder', 'f2')
        f1 = portal.f1
        f2 = portal.f2
        f1.invokeFactory('Document', 'f1d')
        f2.invokeFactory('Document', 'f2d')
        f1d = f1.f1d
        
        f1d.setText('"relativeref":../f2/f2d"', mimetype='text/structured')
        uid = f1d.UID()
        pl.retrieving.retrieveObject(f1d)
        
        db_cont = pl.database.queryLinks(object=[uid])
        self.assertEquals(1, len(db_cont))
        self.assertEqual(portal.absolute_url() + "/f2/f2d",
                         db_cont[0].url)

    def test_urlrecord(self):
        from gocept.linkchecker.utils import urlrecord

        components = ('scheme', 'netloc', 'path', 'parameters', 'query', 'fragment')
        urls = {
            'http://www.gocept.com':
            ['http', 'www.gocept.com', '', '', '', ''],
            'https://www.gocept.com/':
            ['https', 'www.gocept.com', '/', '', '', ''],
            'ftp://user:password@ftp.example.com/pub/barbaz':
            ['ftp', 'user:password@ftp.example.com', '/pub/barbaz', '', '', ''],
            'http://www.example.com:8080/myProject?rot=schwarz&plus=minus#muahaha':
            ['http', 'www.example.com:8080', '/myProject', '', 'rot=schwarz&plus=minus', 'muahaha'],
            '//www.example.com:8080/myProject?rot=schwarz&plus=minus#muahaha':
            ['', 'www.example.com:8080', '/myProject', '', 'rot=schwarz&plus=minus', 'muahaha'],
            'www.example.com/otherProject?rot=schwarz&plus=minus#muahaha':
            ['', '', 'www.example.com/otherProject', '', 'rot=schwarz&plus=minus', 'muahaha'],
            '../myProject?rot=schwarz&plus=minus#muahaha':
            ['', '', '../myProject', '', 'rot=schwarz&plus=minus', 'muahaha'],
            }

        for url, url_ in urls.iteritems():
            url_rec = urlrecord(url)
            url_list = [getattr(url_rec, c, '') for c in components]
            self.assertEqual(url_, url_list)
            self.assertEqual(url, str(urlrecord(url)))

    def test_resolverelative(self):
        from gocept.linkchecker.utils import resolveRelativeLink

        self.loginAsPortalOwner()
        portal = self.portal
        pl = getToolByName(portal, 'portal_linkchecker')
	pl.database.defaultURLPrefix = portal.absolute_url()

        portal.invokeFactory('Folder', 'folder')
        portal.folder.invokeFactory('Document', 'doc')
        portal_id = portal.getId()

        # Something that isn't a URL in the first place is not guaranteed to be
        # handled in a sensible way. For example, this will break things:
        #'something really/weird: which is not @ URL#...'
        urls = {
            # A protocol that should not be touched:
            'mailto:test@example.com': 'mailto:test@example.com',
            # Simple examples of all our protocols:
            'http://www.gocept.com': 'http://www.gocept.com/',
            'https://www.gocept.com': 'https://www.gocept.com/',
            'ftp://ftp.gocept.com': 'ftp://ftp.gocept.com/',
            # Trailing slash without path will be kept, none added.
            'http://www.gocept.com/': 'http://www.gocept.com/',
            # Using a path component:
            'http://www.faqs.org/rfcs/rfc2396.html':
            'http://www.faqs.org/rfcs/rfc2396.html',
            # Trailing slash after a path will be kept, none added.
            'http://www.faqs.org/rfcs/': 'http://www.faqs.org/rfcs/',
            # Using a query:
            'http://www.google.com/search?q=foobar&start=0':
            'http://www.google.com/search?q=foobar&start=0',
            # Omitting just the scheme:
            '//www.google.com/search?q=foobar&start=0':
            'http://www.google.com/search?q=foobar&start=0',
            # Also omitting the host:
            '/search?q=foobar&start=0':
            'http://nohost/search?q=foobar&start=0',
            # a relative URL with regular path elements only:
            'search?q=foobar&start=0':
            'http://nohost/$PORTAL/folder/search?q=foobar&start=0',
            # a relative URL with leading ..:
            '../search?q=foobar&start=0':
            'http://nohost/$PORTAL/search?q=foobar&start=0',
            # a relative URL with leading .. and no-ops:
            './../search?q=foobar&start=0':
            'http://nohost/$PORTAL/search?q=foobar&start=0',
            # a relative URL with a trailing slash:
            './': 'http://nohost/$PORTAL/folder/',
            # a relative URL without a trailing slash:
            '.': 'http://nohost/$PORTAL/folder/',
            # a relative URL with too many leading ..:
            '../../..': 'http://nohost/',
            }

        for before, expected in urls.iteritems():
            after = resolveRelativeLink(before, portal.folder.doc)
            expected = expected.replace('$PORTAL', portal_id)
            self.assertEqual(expected, after)

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LinkCheckerTest))
    return suite



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""Retrieve manager."""

# Zope imports
import transaction
import zope.interface
import gocept.linkchecker.log as log
import zope.lifecycleevent.interfaces
import zope.app.container
import zope.component
from AccessControl import ClassSecurityInfo, getSecurityManager, Unauthorized
from Globals import InitializeClass, PersistentMapping
from OFS.SimpleItem import SimpleItem

import Products.Archetypes.interfaces

# CMF/Plone imports
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName

import gocept.linkchecker.interfaces

def manage_addRetrieveManager(container, id):
    """Add a new retriever manager to a link manager."""
    container._setObject(id, RetrieveManager(id))


class RetrieveManager(SimpleItem):
    """Local registry that manages the mapping between portal_type and 
       specific retrievers for a single plone instance and allows to trigger
       retrieving for an object or the whole site.
    """

    zope.interface.implements(gocept.linkchecker.interfaces.IRetrieveManager)
    security = ClassSecurityInfo()

    def __init__(self, id):
        self.id = id

    ##################
    # IRetrieveManager

    security.declarePublic('retrieveObject')
    def retrieveObject(self, object, online=True):
        # Check for ModifyPortalContent-Permission on context object.
        # This is dangerous, but I think I know what I'm doing.
        sm = getSecurityManager()
        if not sm.checkPermission(ModifyPortalContent, object):
            return
        if (not
            Products.Archetypes.interfaces.IReferenceable.providedBy(object)):
            return
        database = self.getParentNode().database
        database.unregisterObject(object)
        retriever = gocept.linkchecker.interfaces.IRetriever(object, None)
        if retriever is not None:
            links = retriever.retrieveLinks()
            database.registerLinks(links, object, online)

    security.declarePublic('updateObject')
    def updateLink(self, old_link, new_link, object):
        # Check for ModifyPortalContent-Permission on context object.
        # This is dangerous, but I think I know what I'm doing.
        sm = getSecurityManager()
        if not sm.checkPermission(ModifyPortalContent, object):
            raise Unauthorized, "You can't update links for this object."
        retriever = gocept.linkchecker.interfaces.IRetriever(object, None)
        if retriever is not None:
            retriever.updateLink(old_link, new_link)
        self.retrieveObject(object)

    security.declareProtected(ManagePortal, 'retrieveSite')
    def retrieveSite(self):
        """Retrieves the links from all objects in the site."""
        database = self.getParentNode().database
        objects = self.portal_catalog(Language='all')[:2000]
        os_ = len(objects)
        i = 0
        for ob in objects:
            try:
                obj = ob.getObject()
            except Exception, e:
                # Maybe the catalog isn't up to date
                log.logger.debug("Site crawl raised an error for %s: %s" % (ob.getPath(), str(e)))
                continue
            i += 1
            log.logger.debug("Site Crawl Status %s of %s (%s)" % (i, os_, ob.getPath()))
            self.retrieveObject(obj, online=False)
            if not i % 100:
                # Memory optimization
                transaction.savepoint()

    def supportsRetrieving(self, object):
        """Tells if the object is supported for retrieving links."""
        retriever = gocept.linkchecker.interfaces.IRetriever(object, None)
        return bool(retriever)


InitializeClass(RetrieveManager)


@zope.component.adapter(
    zope.app.container.interfaces.IObjectRemovedEvent)
def remove_links(event):
    object = event.object
    try:
        link_checker = getToolByName(object, 'portal_linkchecker').aq_inner
    except AttributeError:
        return
    if not link_checker.active:
        return
    link_checker.database.unregisterObject(object)


@zope.component.adapter(
    zope.lifecycleevent.interfaces.IObjectModifiedEvent)
def update_links(event):
    object = event.object
    try:
        link_checker = getToolByName(object, 'portal_linkchecker').aq_inner
    except AttributeError:
        return
    if not link_checker.active:
        return
    link_checker.retrieving.retrieveObject(object)



# Copyright (c) 2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""CMF link checker tool - link retriever functions"""

import Products.ATContentTypes.content.event
import Products.ATContentTypes.content.link
import gocept.linkchecker.interfaces
import gocept.linkchecker.utils
import zope.component
import zope.interface


class Link(object):
    """Retriever for ATCT Link objects."""

    zope.component.adapts(Products.ATContentTypes.content.link.ATLink)
    zope.interface.implements(gocept.linkchecker.interfaces.IRetriever)

    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        """Finds all links from the object and return them."""
        return [self.context.getRemoteUrl()]

    def updateLink(self, oldurl, newurl):
        """Replace all occurances of <oldurl> on object with <newurl>."""
        if self.context.getRemoteUrl() == oldurl:
            self.context.setRemoteUrl(newurl)


class Event(object):
    """Retriever for ATCT Event objects"""

    zope.component.adapts(Products.ATContentTypes.content.event.ATEvent)
    zope.interface.implements(gocept.linkchecker.interfaces.IRetriever)

    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        """Finds all links from the object and return them."""
        links = [self.context.event_url()]
        for link in gocept.linkchecker.utils.retrieveAllRichTextFields(
                self.context):
            links.append(link)
        return links

    def updateLink(self, oldurl, newurl):
        """Replace all occurances of <oldurl> on object with <newurl>."""
        if self.context.event_url() == oldurl:
            self.context.setEventUrl(newurl)
        gocept.linkchecker.utils.updateAllRichTextFields(
            oldurl, newurl, self.context)


class ATGeneral(object):
    """General retriever for Archetypes that extracts URLs from (rich) text
    fields.
    """

    zope.component.adapts(Products.Archetypes.atapi.BaseContent)
    zope.interface.implements(gocept.linkchecker.interfaces.IRetriever)

    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        """Finds all links from the object and return them."""
        return gocept.linkchecker.utils.retrieveAllRichTextFields(self.context)

    def updateLink(self, oldurl, newurl):
        """Replace all occurances of <oldurl> on object with <newurl>."""
        gocept.linkchecker.utils.updateAllRichTextFields(
            oldurl, newurl, self.context)



# Copyright (c) 2004 gocept gmbh & co. kg
# See also LICENSE.txt
# 

import unittest

from Products.Archetypes.tests.common import *
from Products.Archetypes.tests.utils import *

from Products.Archetypes.tests.test_sitepolicy import makeContent
from Products.CMFCore.utils import getToolByName

import zope.lifecycleevent
import gocept.linkchecker.retrievemanager
from gocept.linkchecker import retrievers

from gocept.linkchecker.tests.base import LinkCheckerTestCase


class LinkCheckerTest(LinkCheckerTestCase):

    def test_event_retriever(self):
        self.loginAsPortalOwner()
        links = ["http://www.asdf.org", "http://www.bauhaus.de",
                 "ftp://www.goodbye.de", "http://www.google.de/img.png"]

        text = " ".join( [ """"asdf":%s\n""" % x for x in links ])
        lc = getToolByName(self.portal, 'portal_linkchecker')

        url = "http://asdf.org"
        event = makeContent(self.portal, portal_type="Event", id="asdf")
        event.edit(event_url=url)
        event.setText(text)

        retriever = retrievers.Event(event)
        links = retriever.retrieveLinks()

        self.assertEqual(len(links), 5)
        self.assertEqual(links[0], url)

        # The general AT retriever doesn't find this URL
        links = retrievers.ATGeneral(event).retrieveLinks()
        self.assertEqual(links[0], "http://www.google.de/img.png")
        self.assertEqual(len(links), 4)

        # The retriever can also update the URL
        retriever.updateLink(url, 'http://foobar.org')
        self.assertEquals(event.event_url(), 'http://foobar.org')

        # If the URL to replace does not match, it doesn't change the URL of
        # the event:
        retriever.updateLink('blubb', 'http://quux.org')
        self.assertEquals(event.event_url(), 'http://foobar.org')

    def test_link_retriever(self):
        self.loginAsPortalOwner()
        lc = getToolByName(self.portal, 'portal_linkchecker')

        url = "http://www.asdf.org/asdf?asdf"
        ob = makeContent(self.portal, portal_type="Link", id="asdf")
        ob.edit(url)

        retriever = retrievers.Link(ob)
        links = retriever.retrieveLinks()

        self.assertEqual(len(links), 1)
        self.assertEqual(links[0], url)

        # The retriever can also update the URL
        retriever.updateLink(url, 'http://www.foobar.org')
        self.assertEquals(ob.getRemoteUrl(), 'http://www.foobar.org')

        # If the URL doesn't match, the link won't be updated
        retriever.updateLink(url, 'http://www.quux.org')
        self.assertEquals(ob.getRemoteUrl(), 'http://www.foobar.org')


    def test_document_html_retriever(self):
        self.loginAsPortalOwner()
        links = ["http://www.asdf.org", "http://www.bauhaus.de",
                 "ftp://www.goodbye.de", "http://www.google.de/img.png"]
        text = " ".join( [ """<a href="%s">asdf</a>""" % x for x in links[:-1] ])
        text += """<img src="%s"/>""" % links[-1]

        lc = getToolByName(self.portal, 'portal_linkchecker')

        url = "http://www.asdf.org/asdf?asdf"
        ob = makeContent(self.portal, portal_type="Document", id="asdf")
        ob.setText(text)

        retriever = retrievers.ATGeneral(ob)
        links_ob = retriever.retrieveLinks()
        self.assertEqualLinks(links, links_ob)

        # The retriever can also update URLs
        retriever.updateLink('http://www.asdf.org', 'http://foo.org')
        self.assertEquals(ob.getText(),
            """<a href="http://foo.org">asdf</a> """
            """<a href="http://www.bauhaus.de">asdf</a> """
            """<a href="ftp://www.goodbye.de">asdf</a>"""
            """<img src="http://www.google.de/img.png" />""")

        # As this goes through various decoding cycles, we need to be careful
        # about encodings:
        ob.setText(u'<a href="asdf.org">\xf6</a>')
        retriever.updateLink('asdf.org', 'bsdf.org')
        # AT converts to utf-8 encoded byte strings internally
        self.assertEquals(ob.getText().decode('utf-8'),
                          u'<a href="bsdf.org">\xf6</a>')

    def test_document_stx_retriever(self):
        self.loginAsPortalOwner()
        links = ["http://www.asdf.org", "http://www.bauhaus.de",
                 "ftp://www.goodbye.de", "http://www.google.de/img.png"]

        text = " ".join( [ """"asdf":%s\n""" % x for x in links ])

        lc = getToolByName(self.portal, 'portal_linkchecker')

        url = "http://www.asdf.org/asdf?asdf"
        ob = makeContent(self.portal, portal_type="Document", id="asdf")
        ob.setText(text, mimetype='text/structured')

        # RichTextRetriever should work just as well.
        links_ob = retrievers.ATGeneral(ob).retrieveLinks()
        self.assertEqualLinks(links, links_ob)

    def assertEqualLinks(self, links, links_ob):
        self.assertEqual(len(links), len(links_ob))
        for x in links_ob:
            self.assert_(x in links)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LinkCheckerTest))
    return suite



# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: setup.py 12305 2008-07-04 14:29:34Z zagy $

from setuptools import setup, find_packages


setup(
    name='gocept.linkchecker',
    version='3.0a2',
    author='gocept gmbh & co. kg',
    author_email='mail@gocept.com',
    description='Check links in your Plone site using a link monitoring server.',
    packages=find_packages('src'),
    package_dir={'':'src'},
    include_package_data=True,
    zip_safe=False,
    license='ZPL 2.1',
    namespace_packages=['gocept'],
    install_requires=['setuptools', 'lxml'],
)



##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""A short script that shortens URLs in a readable fashion to a target length.
"""


import urlparse
                
def shortURL(url, target=50, base=""):

    if len(url) <= target:
        return url 

    if url.startswith(base):
        url = url[len(base):]

    scheme, location, path, parameters, query, fragment = \
        urlparse.urlparse(url)

    for check, filter in filters:
        url_ = urlparse.urlunparse((scheme, location, path, parameters, query, fragment))
        if len(url_) < target:
            break
        if check(scheme, location, path, parameters, query, fragment):
            scheme, location, path, parameters, query, fragment = \
                filter(scheme, location, path, parameters, query, fragment)
            break

    url_ = urlparse.urlunparse((scheme, location, path, parameters, query, fragment))

    if url_ == url:
        return url_

    if url_ <= target:
        return url_

    return shortURL(url_, target)

checkPath = lambda s,l,p,pa,q,f: 1
def plugPath(schema, location, path, parameters, query, fragment):
    path = path.split('/')
    path = [ x for x in path if x != "..." ]
    path = path[:len(path)//2] + ["..."] + path[len(path)//2+1:]
    path = "/".join(path)
    return schema, location, path, parameters, query, fragment

checkParameters = lambda s,l,p,pa,q,f: len(pa) > 3
def plugParameters(schema, location, path, parameters, query, fragment):
    parameters = "..."
    return schema, location, path, parameters, query, fragment
    
checkQuery = lambda s,l,p,pa,q,f: len(q) > 3
def plugQuery(schema, location, path, parameters, query, fragment):
    query = "..."
    return schema, location, path, parameters, query, fragment
    
checkFragment = lambda s,l,p,pa,q,f: len(f) > 3
def plugFragment(schema, location, path, parameters, query, fragment):
    fragment = "..."
    return schema, location, path, parameters, query, fragment


filters = [ (checkFragment, plugFragment), 
            (checkQuery, plugQuery), 
            (checkParameters, plugParameters), 
            (checkPath, plugPath), ]



# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# 
"""CMF link checker tool - basic reports
"""

# Python imports
import time

# Zope imports
import zope.interface
from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from DateTime import DateTime

# CMF/Plone imports
from Products.CMFCore import permissions
from Products.CMFCore.utils import getToolByName

from gocept.linkchecker.interfaces import IBaseReports
from gocept.linkchecker import permissions


def manage_addBaseReports(container, id):
    """Add basic reports to a link manager."""
    container._setObject(id, BaseReports(id))


class BaseReports(SimpleItem):
    """Supplement functions for various basic reports."""

    zope.interface.implements(IBaseReports)
    security = ClassSecurityInfo()

    def _filter_report(self, member, report):
        """Returns a filtered report, which may be None.
        """
        report_states = member.getProperty('lc_notify_details', [])
        if not isinstance(report_states, (list, tuple)):
            return None

        for state in report.keys():
            if state not in report_states:
                del report[state]

        changes_only = member.getProperty('lc_notify_changes_only', False)
        if changes_only:
            last_report = member.getProperty('lc_notify_last_notification',
                                             DateTime())
            for state, links in report.items():
                # remove already reported links (did not change)
                links = [link for link in links
                         if link.lastupdate >= last_report]
                if links:
                    report[state] = links
                    for link in links:
                        if link.laststate in report:
                            # add some extra information
                            report[link.laststate].append(link)
                else:
                    del report[state]
            if not report:
                return None

        return report

    security.declarePublic('GroupedLinksForAuthenticatedMember')
    def GroupedLinksForAuthenticatedMember(self):
        """Returns a list of links aggregated by state for the current user.
        
           The list only features catalog brains from the database.
        """
        user = getSecurityManager().getUser().getId()
        return self.getGroupedLinksFor(user)
    
    security.declareProtected(permissions.ManagePortal,
                              'getGroupedLinksFor')
    def getGroupedLinksFor(self, user_id):

        catalog = getToolByName(self, 'portal_catalog')
        lc = getToolByName(self, 'portal_linkchecker')

        # Find all objects this user is responsible for
        docs = catalog(Creator=user_id, Language='all')
        uids = [ x.UID for x in docs ]
        uids = filter(None, uids)
        links = lc.database.queryLinks(object=uids)

        groups = {'red': [], 'blue': [], 'orange': [], 'grey': [], 'green': []}
        for link in links:
                groups[link.state].append(link)
        return groups


    security.declareProtected(permissions.ManagePortal, "LinksInState")
    def LinksInState(self, state):
        """Returns a list of links in the given state."""

        for link, doc, member in self._document_iterator(state):
            item = {}
            item["url"] = link.url
            item["reason"] = link.reason
            item["lastcheck"] = link.lastcheck
            item["id"] = link.getId
            item["link"] = link.link
            item["document"] = doc
            item["object"] = link.object

            if member is None:
                item["owner_mail"] = ""
                item["owner"] = doc.Creator
            else:
                item["owner_mail"] = member.getProperty("email")
                item["owner"] = member.getProperty("fullname", doc.Creator)

            yield item

    security.declareProtected(permissions.ManagePortal,
                              'ManagementOverview')
    def ManagementOverview(self):
        """Returns a comprehensive management overview over the complete site
        """
        return ManagementReport(self)
        
    def _document_iterator(self, state):
        member_cache = {}
    
        lc = getToolByName(self, 'portal_linkchecker')
        catalog = getToolByName(self, 'portal_catalog')
        pms = getToolByName(self, 'portal_membership')

        _marker = object()

        links = lc.database.queryLinks(state=[state], sort_on="url")
        for link in links:
            doc_uid = link.object
            if not doc_uid:
                continue
            docs = catalog(UID=doc_uid, Language='all')
            for doc in docs:
                creator = doc.Creator
                member = member_cache.get(creator, _marker)
                if member is _marker:
                    member = pms.getMemberById(creator)
                    member_cache[creator] = member
                yield link, doc, member


InitializeClass(BaseReports)


class ManagementReport:

    __allow_access_to_unprotected_subobjects__ = 1

    states = [
        ('green', 'Green'),
        ('orange', 'Orange'),
        ('red', 'Red'),
        ('blue', 'Blue'),
        ('grey', 'Grey'),
    ]

    def __init__(self, context):
        self._context = context
        self._generate_data()

    def _generate_data(self):
        lc = getToolByName(self._context, 'portal_linkchecker')
   
        # get and massage data from the catalog
        links = lc.database.queryLinks()

        link_count_per_document = {}
        link_count_per_state = {'red': 0, 'grey': 0, 'green': 0, 'orange': 0, 'blue': 0}

        for link in links:
            link_count_per_state[link.state] += 1

        # totals
        self.totalLinks = len(links)

        # links in state
        self.links = link_count_per_state
        if links:
            self.linksPct = {}
            for state, count in link_count_per_state.items():
                self.linksPct[state] = float(count) / self.totalLinks * 100
        else:
            self.linksPct = link_count_per_state
